# Modal Lattice Resolution Addon
# Copyright (C) 2022 Yann Terrer aka Tyo-79
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License .


bl_info = {
    "name": "Modal Lattice Resolution",
    "author": "Yann Terrer - Jordan Sautron",
    "version": (0, 3, 4),
    "blender": (4, 2, 0),
    "location": "3D Viewport",
    "description": "Create Fitted Lattice or Tweak Resolution from 3D viewport, 'L' keymap ",
    "category": "Mesh"
}

import bpy
from bpy.props import IntProperty, FloatProperty, FloatVectorProperty, BoolProperty
import blf
from . import preferences
from . import keymaps
from mathutils import Vector
from .shortcuts import draw_callback_px, draw_shortcuts


class MLR_OT_create_lattice(bpy.types.Operator):
    """Create lattice that fit with selected object Size and Rotation"""
    bl_label = "Create Lattice Fitted to Mesh "
    bl_idname = "mlr.create_lattice"

    @classmethod
    def poll(cls, context):
        return context.object.type != 'LATTICE' and context.object and context.mode != "EDIT_MESH"

    def execute(self, context):
        scene = bpy.context.scene.collection
        obj = bpy.context.object
        selected_objects = bpy.context.selected_objects

        if bpy.context.object and obj.type in ["MESH", "CURVE", "TEXT", "SURFACE", "GPENCIL"]:
            if len(selected_objects):
                all_points = []
                for objects in selected_objects:
                    bound_box = objects.bound_box
                    bound_box_points = []
                    world_mat = objects.matrix_world
                    for bound_box_point in bound_box:
                        bound_box_points.append(
                            world_mat @ Vector((bound_box_point[0], bound_box_point[1], bound_box_point[2])))

                    all_points.extend(bound_box_points)

            if len(all_points):
                x, y, z = all_points[0]
                min_x = x
                max_x = x
                min_y = y
                max_y = y
                min_z = z
                max_z = z

                for i in range(1, len(all_points)):
                    x, y, z = all_points[i]
                    min_x = min(x, min_x)
                    max_x = max(x, max_x)
                    min_y = min(y, min_y)
                    max_y = max(y, max_y)
                    min_z = min(z, min_z)
                    max_z = max(z, max_z)

            dim_x = max_x - min_x
            dim_y = max_y - min_y
            dim_z = max_z - min_z

            av_x = (max_x + min_x) / float(2)
            av_y = (max_y + min_y) / float(2)
            av_z = (max_z + min_z) / float(2)

            dimensions = (dim_x, dim_y, dim_z)
            # location = (av_x, av_y, av_z)

            lat = bpy.data.lattices.new("Lattice", )
            lattice = bpy.data.objects.new("Lattice", lat)
            d = lattice.data
            # lattice.show_in_front = True

            if len(selected_objects) == 1:
                lattice.dimensions = obj.dimensions
                for i in range(3):
                    if lattice.scale[i] == 0:
                        lattice.scale[i] = .001
            else:
                lattice.dimensions = dimensions

            if len(selected_objects) == 1:
                lattice.rotation_euler = obj.rotation_euler

            lattice.location = (av_x, av_y, av_z)
            scene.objects.link(lattice)

            for obj in selected_objects:
                if obj.type in ["MESH", "CURVE", "TEXT", "SURFACE"]:
                    mod = obj.modifiers.new("Lattice", 'LATTICE')

                if obj.type == "GPENCIL":
                    mod = obj.grease_pencil_modifiers.new("Lattice", 'GP_LATTICE')

                mod.object = lattice
                lattice.select_set(True)
                bpy.ops.object.parent_set(type='OBJECT', keep_transform=True)

                d.interpolation_type_u = d.interpolation_type_v = d.interpolation_type_w = 'KEY_LINEAR'

                bpy.ops.object.select_all(action='DESELECT')
                bpy.context.view_layer.objects.active = lattice
                lattice.select_set(True)

            bpy.ops.mlr.modal_lattice_res('INVOKE_DEFAULT')

        return {'FINISHED'}


class MLR_OT_modal_lattice_res(bpy.types.Operator):
    """Move an object with the mouse, example"""
    bl_label = "Modal Lattice Resolution "
    bl_idname = "mlr.modal_lattice_res"

    first_mouse_x: IntProperty()
    first_value: FloatProperty()

    user_editmesh_active: FloatVectorProperty(size=4)
    user_edge_select: FloatVectorProperty(size=3)
    user_object_active: FloatVectorProperty(size=3)
    user_wire_edit: FloatVectorProperty(size=3)
    user_wire: FloatVectorProperty(size=3)

    user_u = IntProperty()
    user_v = IntProperty()
    user_w = IntProperty()

    def backup_theme(self):
        theme = bpy.context.preferences.themes[0]
        self.user_editmesh_active = theme.view_3d.editmesh_active
        self.user_edge_select = theme.view_3d.edge_select
        self.user_object_active = theme.view_3d.object_active
        self.user_wire_edit = theme.view_3d.wire_edit
        self.user_wire = theme.view_3d.wire

    def restore_theme(self):
        theme = bpy.context.preferences.themes[0]

        theme.view_3d.editmesh_active = self.user_editmesh_active
        theme.view_3d.edge_select = self.user_edge_select
        theme.view_3d.object_active = self.user_object_active
        theme.view_3d.wire_edit = self.user_wire_edit
        theme.view_3d.wire = self.user_wire

        w = bpy.context.workspace
        w.status_text_set(None)

        # @classmethod
        # def poll(cls, context):
        #     return context.object.type != 'LATTICE' and context.object and context.mode != "EDIT_MESH"

    def modal(self, context, event):
        try:
            d = bpy.context.object.data

            if event.type in {'MIDDLEMOUSE', 'TAB', 'G'}:
                return {'PASS_THROUGH'}

            if event.ctrl and event.type == 'Z':
                return {'PASS_THROUGH'}

            if event.type == 'L':
                d.interpolation_type_u = d.interpolation_type_v = d.interpolation_type_w = 'KEY_LINEAR'

            if event.type == 'B':
                d.interpolation_type_u = d.interpolation_type_v = d.interpolation_type_w = 'KEY_BSPLINE'

            if event.type == 'C':
                d.interpolation_type_u = d.interpolation_type_v = d.interpolation_type_w = 'KEY_CARDINAL'

            if event.type == 'R':
                d.interpolation_type_u = d.interpolation_type_v = d.interpolation_type_w = 'KEY_CATMULL_ROM'

            if event.type == 'P':
                bpy.ops.mlr.resolution_popup('INVOKE_DEFAULT')

            if event.type == 'O' and event.value == 'PRESS':
                bpy.context.object.data.use_outside ^= 1

            if event.type == 'F' and event.value == 'PRESS':
                bpy.context.object.show_in_front ^= 1

            for i in range(2, 10):
                if event.type == f'NUMPAD_{i}':
                    d.points_u = d.points_v = d.points_w = i

            if event.type == 'NUMPAD_0':
                d.points_u = d.points_v = d.points_w = 2

            if event.type == 'NUMPAD_1':
                d.points_u = d.points_v = 2
                d.points_w = 1

            if event.type == 'WHEELUPMOUSE':
                if event.ctrl:
                    d.points_u += 1
                if event.alt:
                    d.points_v += 1
                if event.shift:
                    d.points_w += 1

            if event.type == 'WHEELDOWNMOUSE':
                if event.ctrl:
                    d.points_u -= 1
                if event.alt:
                    d.points_v -= 1
                if event.shift:
                    d.points_w -= 1

            elif event.type in {'LEFTMOUSE', 'BUTTON4MOUSE', 'RET', 'NUMPAD_ENTER'}:

                self.restore_theme()

                bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')

                return {'FINISHED'}

            elif event.type in {'RIGHTMOUSE', 'ESC'}:
                self.restore_theme()

                context.object.data.points_u = self.user_u
                context.object.data.points_v = self.user_v
                context.object.data.points_w = self.user_w

                bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')

                return {'CANCELLED'}

            return {'RUNNING_MODAL'}

        except:
            self.restore_theme()

    def invoke(self, context, event):

        obj = bpy.context.object
        if bpy.context.object and obj.type == "LATTICE" :

            bpy.ops.object.editmode_toggle()
            self.backup_theme()

            self.user_u = context.object.data.points_u
            self.user_v = context.object.data.points_v
            self.user_w = context.object.data.points_w
            bpy.context.object.data.use_outside = True

            blue = preferences.get_preferences().Color_Modal_Lattice
            args = (self, context)
            self._handle = bpy.types.SpaceView3D.draw_handler_add(draw_callback_px, args, 'WINDOW', 'POST_PIXEL')

            bpy.context.preferences.themes[0].view_3d.editmesh_active = blue
            bpy.context.preferences.themes[0].view_3d.edge_select = blue[:3]
            bpy.context.preferences.themes[0].view_3d.object_active = blue[:3]
            bpy.context.preferences.themes[0].view_3d.wire_edit = blue[:3]
            bpy.context.preferences.themes[0].view_3d.wire = blue[:3]

            context.window_manager.modal_handler_add(self)

            w = bpy.context.workspace
            w.status_text_set(draw_shortcuts)

            return {'RUNNING_MODAL'}

        else:
            if context.object and context.mode != "EDIT_MESH":
                bpy.ops.mlr.create_lattice('INVOKE_DEFAULT')
                self.report({'WARNING'}, f"Adding Lattice on '{obj.type}' Object is Not Allowed")
                return {'CANCELLED'}

        return {'CANCELLED'}


class MLR_OT_resolution_popup(bpy.types.Operator):
    '''Lattice Resolution'''
    bl_label = "Lattice Resolution Settings"
    bl_idname = "mlr.resolution_popup"

    def execute(self, context):
        return {'FINISHED'}

    def invoke(self, context, event):
        return context.window_manager.invoke_popup(self, width=250, )

    def draw(self, context, ):
        layout = self.layout
        layout.use_property_split = True

        lat = context.object.data

        col = layout.column()
        col.label(text='Lattice', icon='LATTICE_DATA')
        sub = col.column(align=True)
        sub.prop(lat, "points_u", text="Resolution U")
        sub.prop(lat, "points_v", text="V")
        sub.prop(lat, "points_w", text="W")

        col.separator()

        sub = col.column(align=True)
        sub.prop(lat, "interpolation_type_u", text="Interpolation U")
        sub.prop(lat, "interpolation_type_v", text="V")
        sub.prop(lat, "interpolation_type_w", text="W")

        col.separator()
        col.prop(lat, "use_outside")
        col.separator()
        col.prop_search(lat, "vertex_group", context.object, "vertex_groups")


classes = [
    MLR_OT_modal_lattice_res,
    MLR_OT_create_lattice,
    MLR_OT_resolution_popup,

]

addon_keymaps = []


def register():
    preferences.register()
    keymaps.register()

    for cls in classes:
        bpy.utils.register_class(cls)


def unregister():
    preferences.unregister()
    keymaps.unregister()

    for cls in classes:
        bpy.utils.unregister_class(cls)




